/*
 * DO NOT EDIT THIS FILE - it is generated by Glade.
 */

GtkWidget* create_winMain (void);
GtkWidget* create_papersizeDialog (void);
GtkWidget* create_aboutDialog (void);
GtkWidget* create_zoomDialog (void);
